/**************************************************************

   Program Title:  GenY32 GRBL Port
   
   Program Name:  ser.c	 Version: 1.00
   
   Author:  ALD	  					  Date:  May 10, 2014
   
   Compiled with AVR Studio 6.1.2730
   
   This file contains routines that implement a generic interface
   in both USART serial ports and the USB CDC stack.
   
   Copyright 2014, Tony da Costa  All rights reserved.
   
****************************************************************/
/*
 * Copyright 2014, Tony da Costa  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TONY DA COSTA AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL SOC ROBOTICS, 
 * INC. OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/*
 * ser.c
 *
 * Created: 03/05/2014 6:13:20 PM
 *  Author: tdacosta
 */ 
#include "asf.h"
#include "ser.h"
#include "string.h"

#define XOFF 0x13
#define XON 0x11
#define SER_RX_FLOW

#define SER_USART0_IRQ_PRIORITY    2
#define SER_USART3_IRQ_PRIORITY    2


volatile UINT8 ser_port1TxBuffer[1024];
volatile UINT8 ser_port1RxBuffer[1024];

volatile UINT8 ser_port2TxBuffer[1024];
volatile UINT8 ser_port2RxBuffer[1024];

volatile UINT8 ser_port3TxBuffer[16];
volatile UINT8 ser_port3RxBuffer[512];


typedef struct
{
	volatile UINT8 *txBuffer;
	UINT16 txBufferSize;
	volatile UINT16 txCount;
	UINT16 txPutIndex;
	UINT16 txGetIndex;
	UINT8 txSendXON;
	UINT8 txSendXOFF;
	UINT8 txUnavailable; //For USB
	
	volatile UINT8 *rxBuffer;
	UINT16 rxBufferSize;
	volatile UINT16 rxCount;
	UINT16 rxPutIndex;
	UINT16 rxGetIndex;
	UINT8 rxSwHandshaking;
	UINT16 rxBufferHighMark;
	UINT16 rxBufferLowMark;
	UINT8 rxSwHandshakingXOFFSent;
	
	volatile avr32_usart_t *uartAddr;
	UINT32 baudrate;
	T_SER_CALLBACK_PTR receiveCallBackPtr;
} T_SER_PARAMS;


T_SER_PARAMS ser_portdefs[COM_NUM]=
{
	{ser_port1TxBuffer,sizeof(ser_port1TxBuffer),0,0,0,0,0,0,ser_port1RxBuffer,sizeof(ser_port1RxBuffer),0,0,0,FALSE,(UINT16)(sizeof(ser_port1TxBuffer)*0.8),(UINT16)(sizeof(ser_port1TxBuffer)*0.1),0,&AVR32_USART0,38400,NULL},
	{ser_port2TxBuffer,sizeof(ser_port2TxBuffer),0,0,0,0,0,0,ser_port2RxBuffer,sizeof(ser_port2RxBuffer),0,0,0,FALSE,(UINT16)(sizeof(ser_port2TxBuffer)*0.8),(UINT16)(sizeof(ser_port2TxBuffer)*0.1),0,&AVR32_USART3,38400,NULL},
	{ser_port3TxBuffer,sizeof(ser_port3TxBuffer),0,0,0,0,0,1,ser_port3RxBuffer,sizeof(ser_port3RxBuffer),0,0,0,FALSE,(UINT16)(sizeof(ser_port3TxBuffer)*0.8),(UINT16)(sizeof(ser_port3TxBuffer)*0.1),0,NULL,115200,NULL},
};

void ser_Interrupt(UINT8 comId);


ISR(ser_usart0_irq, AVR32_USART0_IRQ, SER_USART0_IRQ_PRIORITY)
{
	ser_Interrupt(COM1);
}

ISR(ser_usart1_irq, AVR32_USART1_IRQ, SER_USART3_IRQ_PRIORITY)
{
	ser_Interrupt(COM2);
}


void ser_Interrupt(UINT8 comId)
{
	UINT8 ch;
	T_SER_PARAMS *portPtr;
	volatile avr32_usart_t *uartAddr;
	
	if (comId>=COM_NUM) return;
	
	portPtr=ser_portdefs+comId;
	uartAddr=portPtr->uartAddr;
	
	//Receive characters
	while(uartAddr->csr&AVR32_USART_CSR_RXRDY_MASK)
	{
		ch = uartAddr->rhr;
		if (portPtr->receiveCallBackPtr) //Call receive call back function if present
		{
			if ((*(portPtr->receiveCallBackPtr))(ch)) continue; //Don't put the character in the receive buffer if the call back claimed the character
		}
		//Put in buffer or drop if buffer full
		if (portPtr->rxCount<portPtr->rxBufferSize)
		{
			portPtr->rxBuffer[portPtr->rxPutIndex++]=ch;
			if (portPtr->rxPutIndex>=portPtr->rxBufferSize) portPtr->rxPutIndex=0;
			portPtr->rxCount++;
			if ((portPtr->rxSwHandshaking)&&(portPtr->rxCount>portPtr->rxBufferHighMark)&&(!portPtr->txSendXOFF)) //Deal with XON/XOFF flow control
			{
				portPtr->txSendXOFF=1;
				portPtr->txSendXON=0;
				uartAddr->ier = AVR32_USART_IDR_TXRDY_MASK;
			}
		}
	}
	
	//Transmit characters
	if (uartAddr->imr&AVR32_USART_IMR_TXRDY_MASK)
	{
		while (uartAddr->csr&AVR32_USART_CSR_TXRDY_MASK)
		{
			//Send XON/XOFF before any other characters
			if (portPtr->txSendXON)
			{
				portPtr->txSendXON=0;
				uartAddr->thr = XON;
				portPtr->rxSwHandshakingXOFFSent=0;
			}
			else
			if (portPtr->txSendXOFF)
			{
				portPtr->txSendXOFF=0;
				uartAddr->thr = XOFF;
				portPtr->rxSwHandshakingXOFFSent=1;
			}
			if (portPtr->txCount)
			{
				ch=portPtr->txBuffer[portPtr->txGetIndex++];
				if (portPtr->txGetIndex>=portPtr->txBufferSize) portPtr->txGetIndex=0;
				portPtr->txCount--;
				uartAddr->thr = ch;
			}
			else
			{
				uartAddr->idr = AVR32_USART_IDR_TXRDY_MASK;
				break;
			}
		}
	}
}

bool ser_callback_cdc_enable(void)
{	//CDC connected to host
	ser_portdefs[COMUSB].txUnavailable=0;
	return true;
}

void ser_callback_cdc_disable(void)
{	//CDC disconnected from host
	ser_portdefs[COMUSB].txUnavailable=1;
}

void ser_callback_rx_notify(uint8_t port)
{
	UINT8 ch;
	T_SER_PARAMS *portPtr;

	portPtr=ser_portdefs+COMUSB;

	while(udi_cdc_is_rx_ready())
	{
		ch=udi_cdc_getc();
		if (portPtr->receiveCallBackPtr) //Call receive call back function if present
		{
			if ((*(portPtr->receiveCallBackPtr))(ch)) continue; //Don't pass the character to the caller if the call back claimed the character
		}	
		//Put in buffer or drop if buffer full
		if (portPtr->rxCount<portPtr->rxBufferSize)
		{
			portPtr->rxBuffer[portPtr->rxPutIndex++]=ch;
			if (portPtr->rxPutIndex>=portPtr->rxBufferSize) portPtr->rxPutIndex=0;
			portPtr->rxCount++;
		}
	}
}


BOOL ser_Init(void)
{
	BOOL passed=TRUE;
	usart_options_t usart_options;
	int i;

	//Enable I/O pins on modules
	gpio_enable_module_pin(AVR32_USART0_RXD_1_PIN, AVR32_USART0_RXD_1_FUNCTION);
	gpio_enable_module_pin(AVR32_USART0_TXD_1_PIN, AVR32_USART0_TXD_1_FUNCTION);
	gpio_enable_module_pin(AVR32_USART3_RXD_2_PIN, AVR32_USART3_RXD_2_FUNCTION);
	gpio_enable_module_pin(AVR32_USART3_TXD_2_PIN, AVR32_USART3_TXD_2_FUNCTION);

	//Register handlers for the serial ports
	irq_register_handler(ser_usart0_irq, AVR32_USART0_IRQ, SER_USART0_IRQ_PRIORITY);
	irq_register_handler(ser_usart1_irq, AVR32_USART3_IRQ, SER_USART3_IRQ_PRIORITY);

	//Configure and start COM ports
	for (i=0; i<COM_NUM; i++)
	{
		if (ser_portdefs[i].uartAddr)
		{	//USARTS
			//Enable the UART clock
			sysclk_enable_peripheral_clock(ser_portdefs[i].uartAddr);

			//Set the default UART operation
			usart_options.baudrate=ser_portdefs[i].baudrate;
			usart_options.charlength=8;
			usart_options.paritytype=USART_NO_PARITY;
			usart_options.stopbits=USART_1_STOPBIT;
			usart_options.channelmode = USART_NORMAL_CHMODE;
			if (!usart_init_rs232(ser_portdefs[i].uartAddr, &usart_options, sysclk_get_peripheral_bus_hz(ser_portdefs[i].uartAddr))) passed=FALSE;
		
			//Enable RX interrupt
			ser_portdefs[i].uartAddr->ier = AVR32_USART_IER_RXRDY_MASK;
		}
		else
		{	//USB
			udc_start();
		}
	}
	
	return passed;
}


BOOL ser_SetLineParam(UINT8 comId, UINT32 baudrate, UINT8 bits, UINT8 parity, UINT8 stopBits, BOOL swHandshaking)
{
	BOOL passed=TRUE;
	usart_options_t usart_options;

	if (comId>=COM_NUM) return FALSE;

	ser_portdefs[comId].baudrate=baudrate;
	ser_portdefs[comId].rxSwHandshaking=swHandshaking;

	if (ser_portdefs[comId].uartAddr)
	{	//USARTS
		usart_options.baudrate=ser_portdefs[comId].baudrate;
		usart_options.charlength=bits;
		usart_options.paritytype=parity;
		usart_options.stopbits=stopBits;
		usart_options.channelmode = USART_NORMAL_CHMODE;
		if (!usart_init_rs232(ser_portdefs[comId].uartAddr, &usart_options, sysclk_get_peripheral_bus_hz(ser_portdefs[comId].uartAddr))) passed=FALSE;
		ser_portdefs[comId].uartAddr->ier = AVR32_USART_IER_RXRDY_MASK;
	}
	else
	{	//USB
		//Nothing
	}
	return passed;
}

BOOL ser_RegisterReceiveCallback(UINT8 comId, T_SER_CALLBACK_PTR cbPtr)
{
	T_SER_PARAMS *portPtr;
	bool irqenabled;
	
	if (comId>=COM_NUM) return FALSE;
	
	portPtr=ser_portdefs+comId;

	irqenabled=cpu_irq_is_enabled();
	cpu_irq_disable();
	portPtr->receiveCallBackPtr=cbPtr;
	if (irqenabled) cpu_irq_enable();
	return TRUE;
}

BOOL ser_GetChar(UINT8 comId, UINT8 *ch)
{
	T_SER_PARAMS *portPtr;
	bool irqenabled;
	
	if (comId>=COM_NUM) return FALSE;
	if (!ch) return FALSE;
	
	portPtr=ser_portdefs+comId;

	if (portPtr->rxCount)
	{
		*ch=portPtr->rxBuffer[portPtr->rxGetIndex++];
		if (portPtr->rxGetIndex>=portPtr->rxBufferSize) portPtr->rxGetIndex=0;
		irqenabled=cpu_irq_is_enabled();
		cpu_irq_disable();
		portPtr->rxCount--;
		if ((portPtr->rxSwHandshaking)&&(portPtr->rxCount<portPtr->rxBufferHighMark)&&(!portPtr->txSendXON)&&(portPtr->rxSwHandshakingXOFFSent)) //Deal with XON/XOFF flow control
		{
			portPtr->txSendXON=1;
			portPtr->txSendXOFF=0;
			portPtr->uartAddr->ier = AVR32_USART_IDR_TXRDY_MASK;
		}
		if (irqenabled) cpu_irq_enable();
		return TRUE;
	}
	else
	{
		return FALSE;
	}
}

UINT16 ser_Read(UINT8 comId, UINT8 *buffer, UINT16 count)
{
	UINT16 nr;

	if (comId>=COM_NUM) return 0;
	nr=0;
	while (count--)
	{
		if (!ser_GetChar(comId, buffer++))
		{
			return nr;
		}
		nr++;
	}
	return nr;
}

UINT16 ser_GetRxBufferCount(UINT8 comId)
{
	T_SER_PARAMS *portPtr;
	UINT16 count;
	bool irqenabled;
		
	if (comId>=COM_NUM) return 0;
	
	portPtr=ser_portdefs+comId;
	irqenabled=cpu_irq_is_enabled();
	cpu_irq_disable();
	count=portPtr->rxCount;
	if (irqenabled) cpu_irq_enable();
	return count;
}

BOOL ser_FlushRxBuffer(UINT8 comId)
{
	T_SER_PARAMS *portPtr;
	bool irqenabled;

	if (comId>=COM_NUM) return FALSE;

	portPtr=ser_portdefs+comId;

	irqenabled=cpu_irq_is_enabled();
	cpu_irq_disable();
	portPtr->rxCount=0;
	portPtr->rxPutIndex=0;
	portPtr->rxGetIndex=0;
	if ((portPtr->rxSwHandshaking)&&(!portPtr->txSendXON)&&(portPtr->rxSwHandshakingXOFFSent)) //Deal with XON/XOFF flow control
	{
		portPtr->txSendXON=1;
		portPtr->txSendXOFF=0;
		portPtr->uartAddr->ier = AVR32_USART_IDR_TXRDY_MASK;
	}
	
	if (irqenabled) cpu_irq_enable();
	return TRUE;
}

BOOL ser_PutChar(UINT8 comId, UINT8 ch)
{
	T_SER_PARAMS *portPtr;
	bool irqenabled;
	
	if (comId>=COM_NUM) return FALSE;
	
	portPtr=ser_portdefs+comId;

	if (portPtr->uartAddr)
	{	//USART
		//Wait for room in buffer
		while (portPtr->txCount>=portPtr->txBufferSize) {};

		portPtr->txBuffer[portPtr->txPutIndex++]=ch;
		if (portPtr->txPutIndex>=portPtr->txBufferSize) portPtr->txPutIndex=0;
		irqenabled=cpu_irq_is_enabled();
		cpu_irq_disable();
		portPtr->txCount++;
		//Kick the transmit interrupt
		portPtr->uartAddr->ier = AVR32_USART_IER_TXRDY_MASK;
		if (irqenabled) cpu_irq_enable();
		return TRUE;
	}
	else
	{	//USB
		UINT32 startTime;
		startTime=timer_GetTickCount();
		if (portPtr->txUnavailable) return FALSE;
		while(!udi_cdc_is_tx_ready())
		{
			if ((timer_GetTickCount()-startTime)>500) return FALSE;
		}
		return udi_cdc_putc(ch)?TRUE:FALSE;
	}
}

BOOL ser_Write(UINT8 comId, UINT8 *buffer, UINT16 count)
{
	T_SER_PARAMS *portPtr;
	if (comId>=COM_NUM) return FALSE;
	portPtr=ser_portdefs+comId;

	if (portPtr->uartAddr)
	{	//USART
		while (count--)
		{
			if (!ser_PutChar(comId,*buffer++))
			{
				return FALSE;
			}
		}
		return TRUE;
	}
	else
	{	//USB--more efficient to send blocks through USB
		UINT32 bufSpace, written, startTime;
		startTime=timer_GetTickCount();
		while(count)
		{
			if (portPtr->txUnavailable) return FALSE;
			if ((timer_GetTickCount()-startTime)>500) return FALSE;
			bufSpace=udi_cdc_get_free_tx_buffer();
			if (bufSpace)
			{
				if (bufSpace>count) bufSpace=count;
				written=udi_cdc_write_buf(buffer,bufSpace);
				if (written!=bufSpace) return FALSE;
				buffer+=written;
				count-=written;
			}
		}
		return TRUE;		
	}
}

BOOL ser_SendString(UINT8 comId, char *str)
{
	return ser_Write(comId,(UINT8*)str,strlen(str));
}

UINT16 ser_GetTxBufferCount(UINT8 comId)
{
	UINT16 tempu16;
	bool irqenabled;
	T_SER_PARAMS *portPtr;
	
	if (comId>=COM_NUM) return FALSE;
	portPtr=ser_portdefs+comId;
	
	if (portPtr->uartAddr)
	{	//USART
		irqenabled=cpu_irq_is_enabled();
		cpu_irq_disable();
		tempu16=portPtr->txCount;
		if (irqenabled) cpu_irq_enable();
		return tempu16;
	}
	else
	{	//USB
		return 0;
	}
}

